#coding=utf8

# This file is part of CloudControl.
#
# CloudControl is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# CloudControl is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with CloudControl.  If not, see <http://www.gnu.org/licenses/>.

'''
CloudControl CLI Tunnel module
'''

import os
import select
import socket
import threading

from cloudcontrol.cli.printer import Printer
from sjrpc.core import RpcError


class Forward(threading.Thread):
    '''
    Handle a redirection of connections to a local socket to a remote
    destination accross a sjRPC tunnel
    '''

    def __init__(self, rpc, host_id, source_ip, source_port, dest_ip, dest_port):
        # port must be integer
        assert(isinstance(source_port, int))
        assert(isinstance(dest_port, int))
        threading.Thread.__init__(self)
        # no need to keep it if cli exit
        self.daemon = True
        # the underlying connection
        self.rpc = rpc
        # forwarding host
        self.host_id = host_id
        # forwarding infos
        self.source_ip = socket.gethostbyname(source_ip)
        self.dest_ip = socket.gethostbyname(dest_ip)
        self.dest_port = dest_port
        # list of tunnel (multiple forward)
        self.tunnels = {}
        # listenning socket
        self.socket = None
        # Pipe used to signal a stop
        self._stopped, self._stopper = os.pipe()
        # create the listening socket
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.bind((self.source_ip, source_port))
        self.socket.listen(100)
        # get the port, useful is auto generated by OS
        self.source_port = self.socket.getsockname()[1]
        # used to identify
        self.id = "%s:%s" % (self.source_ip, self.source_port)
        # quick printer access
        self.printer = Printer()

    def stop(self):
        '''Close the current forward'''
        os.write(self._stopper, "chiche")

    def run(self):
        '''Thread entry point'''
        try:
            paul = select.poll()
            # register stopped pipe
            paul.register(self._stopped)
            # register listen socket
            paul.register(self.socket)
            while True:
                events = dict(paul.poll())
                # check for stop asked
                if self._stopped in events:
                    break
                new_socket = self.socket.accept()[0]
                # we need to not start tunnel until server have not created
                # the tun in his side. Otherwise, data will be dropped.
                tun = self.rpc.create_tunnel(endpoint=new_socket,
                                             on_shutdown=self.cb_close_tun)
                try:
                    self.printer.debug("Forwarding from %s:%s to %s:%s on host %s"
                                       % (self.source_ip,
                                          self.source_port,
                                          self.dest_ip,
                                          self.dest_port,
                                          self.host_id))
                    label= self.rpc.call("forward", self.host_id,
                                         self.dest_port, self.dest_ip)
                except RpcError as e:
                    self.printer.warn("Unable to forward from %s:%s to %s:%s on host %s"
                                       % (self.source_ip,
                                          self.source_port,
                                          self.dest_ip,
                                          self.dest_port,
                                          self.host_id))
                self.tunnels[tun.label] = tun
        finally:
            self.socket.close()
        # close all tunnels
        for tun in self.tunnels.values():
            tun.shutdown()

    def cb_close_tun(self, tun):
        '''Callback on closing a tunnels related to this forward'''
        self.tunnels.pop(tun.label, 42)
